#
#	FILE:	 Planet_Generator.py
#	AUTHORS:
#	Script code: Nerijus Arlauskas (Nercury)
#	Mac support: AlanH
#	translation into german: Dr. Bakterius
#	Version: 0.47
#-----------------------------------------------------------------------------
#	Copyright (c) 2006 Nercury
#-----------------------------------------------------------------------------

"""
Changelog:

v. 0.47
- Smoother transition from jungle to desert/tundra

v. 0.46
- Both Mac and Windows support in one file. Thanks AlanH for writing code for mac!

v. 0.45
- Oasis

v. 0.44
- Fixed continent distribution bug for high continent count and huge map
- Added more ways how continents can be distributed: you can choose old rule with square on equator, two or three rows, circle. Even more, i added option Smart Selection, which will choose the best distribution method for selected continent count and map size
- Flood plains now have default "Standard" distribution, which makes them appear only on desert, some on tundra and only near fresh water
- Redefined what "natural" continent means. Now it is much more concentrated landmass with possible peninsulas or inland seas, not snake starting in the middle and ending near poles :)
- New landmass generation templates added - Natural, Crab-like, Smooth sides, old natural renamed to Traveling Natural (because it travels alot from starting square)
- New "Smart Random for each continent" shape selection which almost ensures that you will have playable and unpredictable maps with diferent continent and map size selections
- Added a lot of comments to script code
Things to try out:
- New Crab-like continent shape for pangaea map

v. 0.43
- Tested with Warlords: works
- Emergency fix: Grid size function was rewritten incorectly, making map height higher than width ;)

v. 0.42
- Last user choices are now remembered
- Tested with Total Realism 2.0, Fall From Heavens 2.014
- Modified hill intersection function (produces more hills close together)
- Grid size function is now written corectly (compatible with "very huge" map size mode (if there is such in some mod))

v. 0.41
- Some adjustments for humidity
- Fixed some mistakes in options
"""

from CvPythonExtensions import *
import CvUtil
from CvUtil import *
import CvMapGeneratorUtil
from CvMapGeneratorUtil import *
import math
import random
import sys
import gc
import pickle

VERSION = "0_46"

#------------------------------------------------------
"""Dropdown options"""
#------------------------------------------------------
# (title), (script name), (use auto random generation)
selection_titles = [
					["Kontinentanzahl:","continents",True],
					["Kontinentabstand:","continent_separation",True],
					["Kontinentverteilung:","continent_distribution",True],
					["Kontinentform:","continent_shape",True],
					[u"Kontinentgr\xf6\xdfe:","continent_size",True],
					[u"Ozeangr\xf6\xdfe:","ocean",True],
					["Berge:","peaks",False], #do not use auto random generation from array
					[u"H\xfcgel:","hills",False], #do not use auto random generation from array
					["Klima:","variation",True],
					["Feuchtigkeit:","humidity",True],
					["Temperatur:","temperature",True],
					["Schwemmland:","flood_plains",True],
					]
# default choices, or first-time choices (later	only last-time values should be	remembered)
selection_defaults = [2,4,0,2,1,8,7,6,3,3,6,5]
# (selection_description), (value or random	value list), (is included in "random" selection)
# don't be afraid to add, remove or modify values here - it will 
# work as long as you dont do anything stupid ;)
selection_names_and_values = [
                  [ #Continents:
                  ["Zufall","random",False],
                  ["Zufall (1 - 3 Kontinente)",[1,2,3],False],
                  ["Zufall (1 - 5 Kontinente)",[1,2,3,4,5],False],
                  ["Zufall (1 - 7 Kontinente)",[1,2,3,4,5,6,7],False],
                  ["Zufall (2 - 5 Kontinente, Standard)",[2,3,4,5],False],
                  ["Zufall (2 - 7 Kontinente)",[2,3,4,5,6,7],False],
                  ["Zufall (2 - 13 Kontinente)",[2,3,4,5,6,7,8,9,10,11,12,13],False],
                  ["Zufall (2 - 15 Kontinente)",[2,3,4,5,6,7,8,9,10,11,12,13,14,15],False],
                  ["Zufall (3 - 7 Kontinente)",[3,4,5,6,7],False],
                  ["Zufall (3 - 13 Kontinente)",[3,4,5,6,7,8,9,10,11,12,13],False],
                  ["Zufall (3 - 15 Kontinente)",[3,4,5,6,7,8,9,10,11,12,13,14,15],False],
                  ["Zufall (3 - 18 Kontinente)",[3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18],False],
                  ["1 Kontinent",1,True], # True means this will be added to "random" selection
                  ["2 Kontinente",2,True],
                  ["3 Kontinente",3,True],
                  ["4 Kontinente",4,True],
                  ["5 Kontinente",5,True],
                  ["6 Kontinente",6,True],
                  ["7 Kontinente",7,True],
                  ["8 Kontinente",8,True],
                  ["9 Kontinente",9,True],
                  ["10 Kontinente",10,True],
                  ["11 Kontinente",11,True],
                  ["12 Kontinente",12,True],
                  ["13 Kontinente",13,True],
                  ["14 Kontinente",14,False],
                  ["15 Kontinente",15,False],
                  ["16 Kontinente",16,False],
                  ["17 Kontinente",17,False],
                  ["18 Kontinente",18,False],
                  ],
                  [ #Continent separation:
                  ["Zufall (alles)","random",False],
                  ["Zufall (2 - 4)",[2,3,4],False],
                  [u"Mind. 1 Wasserfeld (f\xfcr kleine Karten)",1,True],
                  ["Mind. 2 Wasserfelder",2,True],
                  ["Mind. 3 Wasserfelder (empfohlen)",3,True],
                  [u"Mind. 4 Wasserfelder (sp\xe4ter Kontakt)",4,True],
                  [u"Mind. 5 Wasserfelder (f\xfcr gro\xdfe Karten)",5,True], #actually, forcing this to small map will make it huge
                  ],
                  [ #Continent distribution:
                  ["Sinnvolle Auswahl (Standard)","smart-random",False],
                  [u"Mit Platz am \xc4quator","equator",True],
                  ["Um den Mittelpunkt","center",True],
                  ["In zwei Reihen","2rows",True],
                  ["In drei Reihen","3rows",True],
                  ],
                  [ #Continent shape:
                  ["Zufall","random",False],
                  [u"Sinnvoller Zufall f\xfcr jeden Kontinent","continent-random",True], #makes each continent to use different grow method
                  ["Unvorhersehbar (Standard)","unpredictable",True],
                  [u"Schlangenf\xf6rmig - sehr d\xfcnn","snaky-thin",False], #this generates some very narrow passages
                  [u"Schlangenf\xf6rmig","snaky",True],
                  [u"Schlangenf\xf6rmig - nat\xfcrlich","snaky-natural",True], #tries to keep more round shape (compared to snaky)
                  [u"Weitlaeufig - nat\xfcrlich","traveling-natural",True], #tries to keep more round shape (compared to snaky)
                  [u"Natuerlich","natural",True], #this will be new natural
                  [u"Krebsf\xf6rmig (f\xfcr einen Kontinent)","crab",True],
                  [u"Gleichm\xe4\xdfige (Smooth) Seiten","smooth",False],
                  ],
                  [ #Continent size:
                  ["Zufall","random",False],
                  [u"Gleichgro\xdf","equal",True],
                  [u"Abh\xe4ngig von der Spielerzahl (Standard)","players",True],
                  ["Unvorhersehbar","unpredictable",True], 
                  ],
                  [ #Percentage of ocean:
                  ["Zufall (alles)","random",False],
                  ["Zufall (60% - 70%)",[60,62,64,66,68,70],False],
                  ["Zufall (50% - 60%)",[50,55,60],False],
                  ["Zufall (40% - 70%)",[40,45,50,55,60,65,70],False],
                  [u"Minimale Ozeangr\xf6\xdfe",0,True],
                  [u"Gr\xf6\xdfer 30%",30,True], #sometimes it may not be possible to generate very small ocean size; if algorytm meets dead end or some continent runs out of space, generation stops
                  [u"Gr\xf6\xdfer 40%",40,True],
                  [u"Gr\xf6\xdfer 50%",50,True],
                  [u"Gr\xf6\xdfer 60% (normal)",60,True],
                  [u"Gr\xf6\xdfer 70%",70,True],
                  [u"Gr\xf6\xdfer 80%",80,True],
                  ],
                  [ #Mountains (peaks):
                  ["Zufall (ohne Extreme)","random",False],
                  ["Keine",[0,0,0],True], # [peakStartProbability,peakGrowProbability,maxPeaksInRow]
                  ["Wenige Berge, Gruppen bis zu 3",[0.04,0.5,3],True],
                  ["Wenige Berge, Gruppen bis zu 5",[0.03,0.6,5],True],
                  ["Wenige Berge, Gruppen bis zu 10",[0.03,0.7,10],True],
                  ["Ausgewogene Berge, einzeln",[0.08,0,0],True],
                  ["Ausgewogene Berge, Gruppen bis zu 2",[0.07,0.5,2],True],
                  ["Ausgewogene Berge, Gruppen bis zu 5 (Standard)",[0.067,0.6,5],True],
                  ["Ausgewogene Berge, Gruppen bis zu 7",[0.064,0.7,7],True],
                  ["Ausgewogene Berge, Gruppen bis zu 11",[0.061,0.8,11],True],
                  ["Viele Berge, einzeln",[0.24,0,0],True],
                  ["Viele Berge, Gruppen bis zu 3",[0.2,0.5,3],True],
                  ["Viele Berge, Gruppen bis zu 5",[0.19,0.7,5],True],
                  ["Viele Berge, Gruppen bis zu 10",[0.175,0.8,10],True],
                  ["Viele Berge, Gruppen bis zu 14",[0.16,0.85,14],True],
                  [u"in Gruppen bis zu 100, Spa\xdf oder Schmerz?",[0.1,1,100],False],
                  [u"Extreme Berge, Spa\xdf oder Schmerz?",[0.4,0.95,35],False],
                  ],
                  [ #Hills:
                  ["Zufall","random",False],
                  ["Keine",[0,0,0],True], # [hillStartProbability,hillGrowProbability,maxHillsInRow]
                  [u"Wenige H\xfcgel, einzeln",[0.1,0,0],True],
                  [u"Wenige H\xfcgel, Gruppen bis zu 2",[0.1,0.5,2],True],
                  [u"Wenige H\xfcgel, Gruppen bis zu 5",[0.1,0.55,5],True],
                  [u"Ausgewogene H\xfcgel, einzeln",[0.21,0,0],True],
                  [u"Ausgewogene H\xfcgel, Gruppen bis zu 2 (Standard)",[0.18,0.5,2],True],
                  [u"Ausgewogene H\xfcgel, Gruppen bis zu 5",[0.18,0.55,5],True],
                  [u"Viele H\xfcgel, einzeln",[0.4,0,0],True],
                  [u"Viele H\xfcgel, Gruppen bis zu 2",[0.35,0.5,2],True],
                  [u"Viele H\xfcgel, Gruppen bis zu 5",[0.3,0.55,5],True],
                  [u"Viele H\xfcgel, Gruppen bis zu 9",[0.25,0.7,9],True],
                  ],
                  [ #Climate variation:
                  ["Zufall","random",False],
                  ["Sehr mild",0.8,True],
                  ["Mild",0.6,True],
                  ["Normal",0.4,True],
                  ["Hart",0.3,True],
                  ["Extrem Hart",0.2,True],
                  ],
                  [ #Humidity:
                  ["Zufall","random",False],
                  ["Sehr gering (geringe Vegetation)",0.025,True],
                  ["Gering",0.2,True],
                  ["Normal",0.4,True],
                  ["Feucht",0.6,True],
                  ["Hoch",0.78,True],
                  ["Sehr hoch (hohe Vegetation)",0.95,True],
                  ],
                  [ #Temperature:
                  ["Zufall (alles)","random",False],
                  ["Zufall (normal)",[15,20,25,30,35,40,45,50,55,60],False],
                  ["Extrem kalt",-9,True],
                  ["Sehr kalt (Tundra in der Mitte)",10,True],
                  [u"Kalt (W\xe4lder in der Mitte)",20,True],
                  [u"Kalt (leichter Dschungel, keine W\xfcsten)",30,True],
                  ["Normal",40,True],
                  [u"Warm (mehr W\xfcsten / Dschungel)",55,True],
                  [u"Hei\xdf",70,True],
                  [u"Extrem hei\xdf",85,True],
                  ],
                  [ #Flood plains: bellow zero - standart distribution method, ranging from -1 = no floods to 0.001 flood hell, above zero - floods can spread, 0 = lowest spread level
                  ["Zufall (alles)","random",False],
                  ["Zufall (wenige/normal/viele)",[-1,0,1],False],
                  ["Standard - wenige",-0.9,True],
                  ["Standard - normal",-0.75,True],
                  ["Standard - viele",-0.3,True],
                  [u"Auf allen Gel\xe4ndearten - wenige, nur am S\xfc\xdfwasser",0,True],
                  [u"Auf allen Gel\xe4ndearten - normal, auch am Salzwasser",1,False],
                  [u"Auf allen Gel\xe4ndearten - viel (weitverbreitet)",3,False],
                  [u"Auf allen Gel\xe4ndearten - extrem viel",7,False],
                  ],
                  ]

continentCount = 1 #initial value
continentSeparation = 3 #initial value
climateVariation = 0.4
climateHumidity = 0.5
climateTemperature = 40

world = None

landTypes = [PlotTypes.PLOT_LAND,PlotTypes.PLOT_HILLS,PlotTypes.PLOT_PEAK]

generationStarted = False

def getSelValue(name):
	"""Returns None if not found"""
	cgc = CyGlobalContext()
	cymap = CyMap()
	dice = cgc.getGame().getMapRand()
	optionIndex = 0
	found = False
	for optionIndex in range(len(selection_titles)):
		title = selection_titles[optionIndex][1]
		if title == name:
			found = True
			break
	if found:
		print "---------- chosen option:",optionIndex
		index = int(cymap.getCustomMapOption(optionIndex))
		print "---------- selected value index:",index
		if index < len(selection_names_and_values[optionIndex]):
			print "---------- selected value name:",selection_names_and_values[optionIndex][index][0]
			val = selection_names_and_values[optionIndex][index][1]
			print "---------- selected actual value:",selection_names_and_values[optionIndex][index][1]
			print "---------- val type",type(val)
			if val == "random":
				colInd = []
				for i in range(len(selection_names_and_values[optionIndex])):
					if selection_names_and_values[optionIndex][i][2]:
						colInd.append(i)
				print colInd
				drop = dice.get(len(colInd),"Python - EarthGen: Random coice for menu selection")
				print "drop:",colInd[drop]
				val = selection_names_and_values[optionIndex][colInd[drop]][1]
				print "---------- selected random value:",val
			if selection_titles[optionIndex][2] and isinstance(val,list): #if use auto random seletion and value is array
				drop = dice.get(len(val),"Python - EarthGen: Random choice for menu selection")
				print "Random menu option, drop:",drop
				return val[drop]
			else:
				return val
		else:
			val = selection_names_and_values[optionIndex][selection_defaults[optionIndex]][1]
	else:
		return None
#------------------------------------------------------
"""MAP SETUP FUNCTIONS"""
#------------------------------------------------------

def	getDescription():
	"""
	A map's	Description	is displayed in	the	main menu when players go to begin a game.
	For	no description return an empty string.
	"""
	return unicode("Map generator, capable of creating very unpredictable continent shapes, while keeping continent size in control and balance.")

def	getGridSize(argsList):
	"""Given an	argument of	[worldSize], where worldSize is	of type	WorldSizeTypes.
	Can	be overridden to return	a (width, height) tuple	representing the number	of terrain cells
	the	world should have. (a terrain cell is a	chunk of 4x4 plots)"""
	#[worldSize] = argsList
	global generationStarted

	saveMapOptionDefaults()

	generationStarted = True
	if (argsList[0] == -1): # (-1,) is passed to function on loads ; thanks smart map ;)
		return (-1,-1)
	global continentCount
	global continentSeparation
	continentCount = getSelValue("continents")
	if continentCount is None:
		continentCount = 1
	continentSeparation = getSelValue("continent_separation")
	if continentSeparation is None:
		continentSeparation = 3

	distribution = getContinentDistribution()
	minWidthDivision = 1
	if distribution == "2rows":
		minWidthDivision = 2
	elif distribution == "3rows":
		minWidthDivision = 3
	
	cgc = CyGlobalContext()

	sizeDuel = cgc.getInfoTypeForString("WORLDSIZE_DUEL")
	sizeTiny = cgc.getInfoTypeForString("WORLDSIZE_TINY")
	sizeSmall = cgc.getInfoTypeForString("WORLDSIZE_SMALL")
	sizeStandart = cgc.getInfoTypeForString("WORLDSIZE_STANDARD")
	sizeLarge = cgc.getInfoTypeForString("WORLDSIZE_LARGE")
	sizeHuge = cgc.getInfoTypeForString("WORLDSIZE_HUGE")

	map = cgc.getMap()
	worldSize = map.getWorldSize()
	if worldSize <= sizeDuel: #WORLDSIZE_DUEL:
		iW = 40
		iH = 24
	elif worldSize <= sizeTiny: #WORLDSIZE_TINY:
		iW = 48
		iH = 28
	elif worldSize <= sizeSmall: #WORLDSIZE_SMALL:
		iW = 52
		iH = 32
	elif worldSize <= sizeStandart: #WORLDSIZE_STANDARD:
		iW = 64
		iH = 40
	elif worldSize <= sizeLarge: #WORLDSIZE_LARGE:
		iW = 84
		iH = 52
	elif worldSize <= sizeHuge: #WORLDSIZE_HUGE:
		iW = 128
		iH = 80
	else: #maybe very huge?
		iW = 140
		iH = 88

	fW = iW / 4
	fH = iH / 4
	#check to see if there is enough place for continents to start growing
	#if not, increase grid width
	if distribution == "center":
		perimeter = (continentSeparation + 3) * continentCount
		radius = perimeter / (2 * 3.141592)
		width_multiplier = 1.5
		required_width = width_multiplier * radius * 4
		if required_width > fW * 4:
			fnW = required_width / 4
			fH = int(fH * fnW / float(fW) * 4) / 4
			fW = fnW
	else:
		if ((continentSeparation * 2) + 1) * continentCount / minWidthDivision > fW * 4:
			fnW = (((continentSeparation * 2) + 1) * continentCount / minWidthDivision + continentCount) / 4
			fH = int(fH * fnW / float(fW) * 4) / 4
			fW = fnW
	return (fW, fH)
	# return (10, 6) # tiny
	#return (32, 20) # huge

# array wrapper with binary search optimization
# usage:
# - creating array:
# a = bArray()
# - adding data:
# a[2] = "data1"
# b[5] = "data2"
# - retrieving data:
# if a[2] != None:
#    print a[2]
# - length:
# print len(a)
# - removing item:
# a.pop(itemIndex)
# - removing item by key
# a.remove(5)
# - taking random item from array and removing it at the same time:
# print a.random(dice)
# - taking random item from array and not removing it:
# print a.randomItem(dice)
# - reading array data:
# for data in a:
#    print data
class bArray:
	def __init__(self):
		self._data = []
	
	def __setitem__(self, power, newValue):
		index, value = self._search(power,0,len(self._data)-1)
		if value == None:
			self._data.insert(index, [power, newValue])
		else:
			self._data[index][1] = newValue
	
	def __getitem__(self, power):
		index, value = self._search(power,0,len(self._data)-1)
		return value
	
	def __len__(self):
		return len(self._data)
	
	def pop(self, index = None):
		if index == None:
			return self._data.pop()
		else:
			return self._data.pop(index)

	def remove(self, power):
		index, value = self._search(power,0,len(self._data)-1)
		if value != None:
			self._data.pop(index)
			return True
		else:
			return False

	def random(self, dice):
		if len(self._data) == 0:
			return None
		else:
			item = dice.get(len(self._data),"Python - EarthGen: Random bArray item with removing.")
			value = self._data[item][1]
			self._data.pop(item)
			return value

	def randomItem(self, dice):
		if len(self._data) == 0:
			return None
		else:
			item = dice.get(len(self._data),"Python - EarthGen: Random bArray item without removing.")
			value = self._data[item][1]
			return value

	def _search(self,value,fromI,toI):
		left = fromI
		right = toI
		while left <= right:
			mid = (right-left) / 2 + left
			if value > self._data[mid][0]:
				left = mid + 1
			elif value < self._data[mid][0]:
				right = mid-1
			else:
				return mid, self._data[mid][1]
		return left, None

	def __iter__(self):
		self.index = -1
		return self

	def next(self):
		if self.index == len(self._data) - 1:
			raise StopIteration
		self.index += 1
		return self._data[self.index][1]

class probabilityArray:
	def __init__(self):
		self._data = []
		self._sum = 0.0
		self._maxpower = bArray()
		self._maxpowerindex = -1

	def __setitem__(self, power, newValue):
		if power > 0:
			self._data.append([power, newValue])
			self._sum += float(power)
			if power > self._maxpower:
				self._maxpower = power
				self._maxpowerindex = len(self._data) - 1
		
	def __len__(self):
		return len(self._data)
	
	def random(self, dice):
		if len(self._data) > 0:
			item = getRand(dice) * self._sum
			counted = 0
			foundindex = -1
			for index in range(len(self._data)):
				counted += self._data[index][0]
				if counted >= item:
					foundindex = index
					break
			if foundindex >= 0:
				power, value = self._data[foundindex]
				self._data.pop(foundindex)
				self._sum -= power
				return value
			else:
				if self._maxpowerindex != -1:
					power, value = self._data[self._maxpowerindex]
					self._data.pop(self._maxpowerindex)
					self._sum -= power
					self._maxpower = 0.0
					for index in range(len(self._data)):
						newpower, newvalue = self._data[index]
						if newpower > self._maxpower:
							self._maxpower = newpower
							self._maxpowerindex = index
					return value
		return None

	def randomItem(self, dice):
		if len(self._data) > 0:
			item = getRand(dice) * self._sum
			counted = 0
			foundindex = -1
			for index in range(len(self._data)):
				counted += self._data[index][0]
				if counted > item:
					foundindex = index
					break
			if foundindex >= 0:
				power, value = self._data[foundindex]
				return value
			else:
				if self._maxpowerindex != -1:
					return self._data[self._maxpowerindex][1]
		return None

# multi dimensional array wrapper
# usage:
# - creating array:
# a = mdArray([5,9],0) #creates 59 array with data:0 in every cell
# b = mdArray([4,2,6],"nodata") #creates 426 array with data:"nodata" in every cell
# - changing data:
# a[0,1] = 5 #change 01 cell to 5
# b[3,1,5] = "some data" #change 315 cell to "some data"
# - also:
# print a[0,1]
# print len(a)
# - changing data and retrieve offset:
# offset = a.getOffset([1,1])
# a._data[offset] = 6
class mdArray:
	"multi dimensional array handler"
	def __init__(self, dimensions, initial = None):
		dimlen = len(dimensions)
		self._dimensions = [0] * dimlen
		self._factors = [0] * dimlen
		product = 1
		i = dimlen - 1
		while i >= 0:
			self._dimensions[i] = dimensions[i]
			self._factors[i] = product
			product *= self._dimensions[i]
			i -= 1
		self._data = [initial] * product

	def getOffset(self, indices):
		if len(indices) != len(self._dimensions):
			raise IndexError
		offset = 0
		for i, dim in enumerate(self._dimensions):
			if indices[i] < 0 or indices[i] >= dim:
				raise IndexError
			offset += self._factors[i] * indices[i]
		return offset
	
	def __getitem__(self, indices):
		return self._data[self.getOffset(indices)]
	
	def __setitem__(self, indices, value):
		self._data[self.getOffset(indices)] = value

	def get(self):
		return self._data
	
	def __len__():
		return len(self._data)

#for saving grid data, wraps in w
class worldArray(mdArray):
	def __init__(self, dimensions, initial = None):
		if len(dimensions) != 2: #only allow 2 dimensions
			raise IndexError
		dimensions_copy = [dimensions[1],dimensions[0]] #reverse dimensions (required to pass plot types corectly)
		dimlen = len(dimensions_copy)
		self._dimensions = [0] * dimlen
		self._factors = [0] * dimlen
		product = 1
		i = dimlen - 1
		while i >= 0:
			self._dimensions[i] = dimensions_copy[i]
			self._factors[i] = product
			product *= self._dimensions[i]
			i -= 1
		self._data = [initial] * product

	def getOffset(self, indices):
		if len(indices) != 2: #only allow 2 indices
			raise IndexError

		indices_copy = [indices[1],indices[0]] #reverse indices (required to match dimensions)

		if len(indices_copy) != len(self._dimensions) or len(indices_copy) < 2:
			raise IndexError
		
		# make sure array wraps if out of bounds
		while indices_copy[0] < 0:
			indices_copy[0] += self._dimensions[0]
		while indices_copy[0] > self._dimensions[0] - 1:
			indices_copy[0] -= self._dimensions[0]
		
		while indices_copy[1] < 0:
			indices_copy[1] += self._dimensions[1]
		while indices_copy[1] > self._dimensions[1] - 1:
			indices_copy[1] -= self._dimensions[1]

		offset = 0
		for i, dim in enumerate(self._dimensions):
			if indices_copy[i] < 0 or indices_copy[i] >= dim:
				raise IndexError
			offset += self._factors[i] * indices_copy[i]
		return offset
	
	def getHeight(self):
		return self._dimensions[0]

	def getWidth(self):
		return self._dimensions[1]
	
# snaky continent generation algorythm
class TileBuilder:
	def __init__(self,world,tileType,builderTiles,activeTiles,growCount,continentSeparation,targetSize):
		self.cgc = CyGlobalContext()
		self.dice = self.cgc.getGame().getMapRand()
		self.map = self.cgc.getMap()
		self.world = world
		self.tileType = [tileType] # what tile type we are building
		if isinstance(builderTiles,list):
			self.builderTiles = builderTiles
		else:
			self.builderTiles = [builderTiles] # what tiles spawn new tiles
		if isinstance(activeTiles,list):
			self.activeTiles = activeTiles
		else:
			self.activeTiles = [activeTiles] # how many tiles are active for becomming "builder" tiles in case builders extinct
		if isinstance(growCount,list):
			self.growCount = growCount
		else:
			self.growCount = [growCount] # how much tiles one builder tile tries to spawn per life
		self.continentSeparation = continentSeparation # how far same type of tiles (but not active) must be
		self.baseTerrain = [] #add plot types to this array to make tile builder build only on them
		self.targetSize = targetSize
	
	def start(self, x, y):
		self.progress = 0
		self.world[x,y] = self.tileType[0]
		offset = self.world.getOffset([x,y])
		self.builders = bArray() # for storing sorted builder tiles for fast access
		self.builders[offset] = [x, y]
		self.active = bArray() # for storing sorted active tiles for fast access
		self.active[offset] = [x, y, self.getPrValue(self.activeTiles) / self.getPrValue(self.growCount)]
		self._growInd = 0
		self.builderList = [] # for storing sequence in which builder tiles were created
		self.builderList.append(offset)
		self.possibleTiles = bArray() # for storing possible grow tile list
	
	def getPrValue(self, prArray):
		if len(prArray) == 1:
			return int(prArray[0])
		elif self.progress >= self.targetSize:
			return int(prArray[len(prArray) - 1])
		else:
			scale = self.progress / float(self.targetSize)
			onepiece = 1 / float(len(prArray))
			piecesum = onepiece
			index = 0
			while scale > piecesum:
				piecesum += onepiece
				index += 1
			if index >= len(prArray) - 1:
				return int(prArray[len(prArray) - 1])
			else:
				piecesum -= onepiece
				lscale = (scale - piecesum) / float(onepiece)
				return int(prArray[index] + (prArray[index + 1] - prArray[index]) * lscale)
	
	def grow(self):
		world = self.world
		result = True

		builderTiles = self.getPrValue(self.builderTiles)
		activeTiles = self.getPrValue(self.activeTiles)
		growCount = self.getPrValue(self.growCount)

		# increase index (because only one tile per function)
		self._growInd += 1
		# pradzioj rasti langelius aplink
		if self._growInd == 1 and len(self.possibleTiles) < 8: # if there is shortage of possible tiles, remake possible tile list
			self.possibleTiles = self._getPossibleGrowTiles()
		
		# try to find more tiles if there is no more
		#fast method first:
		if len(self.possibleTiles) == 0:
			self.builders = bArray()
			c = 0
			for j in range(builderTiles * 2):
				tile = self.active.randomItem(self.dice)
				x, y, d = tile
				c = self._checkPosGrow(self.possibleTiles,x,y)
				if (c > 0):
					if len(self.builders) < builderTiles:
						offset = world.getOffset([x,y])
						self.builders[offset] = [x, y]
						self.builderList.append(offset)
					else:
						break

		#extreme method (slow as hell):
		if len(self.possibleTiles) == 0:
			for tile in self.active:
				x, y, d = tile
				c = self._checkPosGrow(self.possibleTiles,x,y)
				if (c > 0):
					if len(self.builders) < builderTiles:
						offset = world.getOffset([x,y])
						self.builders[offset] = [x, y]
						self.builderList.append(offset)

		# grow one tile
		if len(self.possibleTiles) > 0:
			x, y = self.possibleTiles.random(self.dice) # take one tile from possible tile array
			offset = world.getOffset([x, y])
			self.active[offset] = [x, y, activeTiles / growCount]
			world[x, y] = self.tileType[0]
			#add tile as possible builder
			self.builders[offset] = [x, y]
			self.builderList.append(offset)
		else:
			return False

		# readjusting builders
		while len(self.builders) > builderTiles:
			offset = self.builderList.pop(0)
			self.builders.remove(offset)

		if self._growInd == growCount:
			self._growInd = 0
			#decrease active tiles
			i = len(self.active)
			while i > 0:
				i -= 1
				if self.active._data[i][1][2] < 1:
					self.active.pop(i)
				else:
					self.active._data[i][1][2] -= 1

		self.progress += 1

		return result
	
	def _intersect(self, x, y):
		dice = self.dice

		cs = self.continentSeparation
		
		r = range(cs * 2)
		for i in r:
			ax = x - cs + i
			ay = y + cs
			offset = self.world.getOffset([ax, ay])
			if (not self.active[offset]) and self.world[ax, ay] in self.tileType:
				return True
		for i in r:
			ax = x + cs
			ay = y + cs - i
			offset = self.world.getOffset([ax, ay])
			if (not self.active[offset]) and self.world[ax, ay] in self.tileType:
				return True
		for i in r:
			ax = x + cs - i
			ay = y - cs
			offset = self.world.getOffset([ax, ay])
			if (not self.active[offset]) and self.world[ax, ay] in self.tileType:
				return True
		for i in r:
			ax = x - cs
			ay = y - cs + i
			offset = self.world.getOffset([ax, ay])
			if (not self.active[offset]) and self.world[ax, ay] in self.tileType:
				return True
		return False
	
	# dont ask me to explain anything here, just keep reading and enjoy :D
	def _checkPosGrowMatrix(self,matrix,list,x,y):
		world = self.world
		c = 0
		h = len(matrix)
		w = len(matrix[0])
		dh = (h - 1) / 2
		dw = (w - 1) / 2
		for ax in range(w):
			for ay in range(h):
				cell = matrix[ay][ax]
				if cell > 0:
					px = x + (ax - dw)
					py = y + (ay - dh)
					if (len(self.baseTerrain) == 0):
						wtt = world[px,py]
						if py > y:
							if py < world._dimensions[0]-self.continentSeparation-1 and not wtt in self.tileType and not self._intersect(px,py):
								list[world.getOffset([px,py])] = [px,py]
								c += 1
						elif py < y:
							if py > self.continentSeparation and not wtt in self.tileType and not self._intersect(px,py):
								list[world.getOffset([px,py])] = [px,py]
								c += 1
						else:
							if not wtt in self.tileType and not self._intersect(px,py):
								list[world.getOffset([px,py])] = [px,py]
								c += 1
					else:
						wtt = world[px,py]
						if py > y:
							if py < world._dimensions[0]-self.continentSeparation-1 and wtt in self.baseTerrain and not self._intersect(px,py):
								list[world.getOffset([px,py])] = [px,py]
								c += 1
						elif py < y:
							if py > self.continentSeparation and wtt in self.baseTerrain and not self._intersect(px,py):
								list[world.getOffset([px,py])] = [px,py]
								c += 1
						else:
							if wtt in self.baseTerrain and not self._intersect(px,py):
								list[world.getOffset([px,py])] = [px,py]
								c += 1
		return c
				
	def _checkPosGrow(self,list,x,y):
		world = self.world
		dice = self.dice
		c = 0
		iH = world.getHeight()
		#matrix of what tiles to check around builder tiles. 1 - available for building, 0 - not
		flipUp = False
		flipDown = False
		if y > iH / 2: #get chance to flip up
			scale = (iH - y) / (iH / 2.0) # 0 - equator, 1 - pole
			if getRand(dice) < scale:
				flipUp = True
		elif y < iH / 2:
			scale = 1 - y / (iH / 2.0) # 0 - equator, 1 - pole
			if getRand(dice) < scale:
				flipDown = True
			
		default_matrix = [
			[1,1,1],
			[1,0,1],
			[1,1,1],
		]
		if flipUp:
			default_matrix = [
				[1,1,1],
				[1,0,1],
				[1,0,1],
			]
		elif flipDown:
			default_matrix = [
				[1,0,1],
				[1,0,1],
				[1,1,1],
			]
		c += self._checkPosGrowMatrix(default_matrix, list, x, y)

		return c

	def _getPossibleGrowTiles(self):
		world = self.world
		result = bArray()
		rem_ind = []
		i = 0
		for item in self.builders:
			x, y = item
			c = self._checkPosGrow(result,x,y)
			if c == 0:
				rem_ind.append(i)
			i += 1
		# remove unecessary builder tiles
		x = len(rem_ind)
		while x > 0:
			x -= 1
			self.builders.pop(x)
		return result

def	generateRandomMap():
	"""Can be overridden if	you	want to	manually control the sequence in which plot	types, heights,	and	terrain	are	generated.
	
	WARNING: If	you	override this function,	then none of the following functions will be automatically called for you:
		generateTerrain()
		generatePlotTypes()
	
	However, unless	you	have some specific reason to change	the	sequence of	map	generation,	you	should probably	just override
	one	or more	of the functions listed	above.
	"""
	print "--------- Generate random map"
	CyPythonMgr().allowDefaultImpl()

def initReplacements():
	# 0 - nothing
	# 1 - water
	# 2 - land
	rep = []
	orig = [
		[0,2,0],
		[2,1,2],
		[0,2,0],
		]
	final = [
		[2],
		]
	rep.append([orig,final])
	orig = [
		[0,2,2,0],
		[2,1,1,2],
		[0,2,2,0],
		]
	final = [
		[2,2],
		]
	rep.append([orig,final])
	orig = [
		[0,2,0],
		[2,1,2],
		[2,1,2],
		[0,2,0],
		]
	final = [
		[2],
		[2],
		]
	rep.append([orig,final])
	return rep

def landR(i):
	if i == 0:
		return None
	elif i == 1:
		return PlotTypes.PLOT_OCEAN
	elif i == 2:
		return PlotTypes.PLOT_LAND

# called by Firaxis - expects to get array of plot types
def generatePlotTypes():
	global generationStarted
	global selection_titles

	global world # our world array

	global continentCount
	global continentSeparation

	cgc = CyGlobalContext()
	map = cgc.getMap()

	if generationStarted:
		print "=============== receives latitudes and map size =============== "
	else:
		print "=============== no lattude or map size change, regeneration =============== "

	#print "generate plot types"
	dice = cgc.getGame().getMapRand() #dice is Firaxis' random number generator
	players = cgc.getGame().countCivPlayersEverAlive()
	iW = map.getGridWidth()
	iH = map.getGridHeight()
	#print "----------------- generating world",
	oceanPercentage = getSelValue("ocean")
	if oceanPercentage is None or oceanPercentage > 100: #ensure we don't make something bad if wrong settings exist
		oceanPercentage = 70
	continentShape = getSelValue("continent_shape") #string value
	if continentShape is None: #if something bad with settings, jump to normal
		continentShape = "natural"
	continentSize = getSelValue("continent_size")
	if continentSize is None:
		continentSize = "players"

	# there is list of templates which are used for various continent shape types
	# there are three parameters in this array (separated by commas): [builder tiles, active tiles, grow count]

	#first type of natural shape:
	#this happens to be like snaky
	tbsSnakyNatural = [2, iW * iH / 8, 2] # iW * iH / 8 = MapWidth * MapHeight / 8 = active tiles

	#second type of natural shape:
	#more thick than first natural
	tbsTravelingNatural = [4, iW * iH / 10, 4] # iW * iH / 10 = MapWidth * MapHeight / 10 = active tiles

	#new2 - spreads equally to all directions, posibility of peninsulas and inland seas
	tbsNatural = [[7,6,4], [iW * iH / 12, iW * iH / 6], [4,4,4,2,1]]

	#new1 - spreads equaly to all directions, can huge peninsulas happen
	tbsSmooth = [15, iW * iH / 3, 4]

	#new3 - spreads equally to all directions, has huge landmass in the middle and many peninsulas, low probability of inside lakes
	sn = iW * iH / 3
	if sn < 40:
		sn = 40
		tbsCrabLike = [20, sn, 2]	
	else:
		tbsCrabLike = [20, sn, 2]
	
	#snaky template:
	#when generating snaky continent, we must ensure active tile count is not too low, so we check for that
	sn = iW * iH / 60
	if sn < 40:
		sn = 40
	tbsSnaky = [3, sn, 2]

	#snaky-thin template:
	#this is extreemly thin
	sn = iW * iH / 50
	if sn < 40:
		sn = 40
	tbsSnakyThin = [2, sn, 1]
	
	# if random for each continent is chosen
	if continentShape == "continent-random":
		shapeRandom = True 
		if iW < 64: # < standart
			if continentCount <= 2:
				randomList = [tbsSnaky,tbsSnakyNatural,tbsTravelingNatural]
			elif continentCount <= 5:
				randomList = [tbsSnaky,tbsSnakyNatural,tbsTravelingNatural,tbsNatural]
			else:
				randomList = [tbsSnaky,tbsSnakyNatural,tbsTravelingNatural,tbsNatural,tbsCrabLike]
		elif iW < 128: # < huge
			if continentCount <= 2:
				randomList = [tbsSnakyNatural,tbsTravelingNatural,tbsNatural,tbsCrabLike]
			elif continentCount <= 5:
				randomList = [tbsSnaky,tbsSnakyNatural,tbsTravelingNatural,tbsNatural,tbsCrabLike]
			else:
				randomList = [tbsSnaky,tbsSnakyNatural,tbsTravelingNatural,tbsCrabLike]
		else: # >= huge
			if continentCount == 1:
				randomList = [tbsTravelingNatural,tbsSmooth,tbsNatural,tbsCrabLike]
			elif continentCount == 2:
				randomList = [tbsTravelingNatural,tbsSmooth,tbsNatural,tbsCrabLike]
			elif continentCount <= 5:
				randomList = [tbsSnakyNatural,tbsTravelingNatural,tbsNatural,tbsCrabLike]
			else:
				randomList = [tbsSnaky,tbsSnakyNatural,tbsTravelingNatural,tbsNatural,tbsCrabLike]
	else: #otherwise we select template
		shapeRandom = False 
		if continentShape == "snaky-natural":
			shapeData = tbsSnakyNatural
		elif continentShape == "traveling-natural":
			shapeData = tbsTravelingNatural
		elif continentShape == "unpredictable":
			shapeRandom = True
			randomList = []
			for i in range(continentCount):
				randomList.append([[2 + getRand(dice) * 3, 2 + getRand(dice) * 3, 2 + getRand(dice) * 3], [iW * iH / (5 + getRand(dice) * 10), iW * iH / (5 + getRand(dice) * 10), iW * iH / (5 + getRand(dice) * 10)], [2 + getRand(dice) * 2, 2 + getRand(dice) * 2]])			
		elif continentShape == "snaky":
			shapeData = tbsSnaky
		elif continentShape == "snaky-thin":
			shapeData = tbsSnakyThin
		elif continentShape == "smooth":
			shapeData = tbsSmooth
		elif continentShape == "natural":
			shapeData = tbsNatural
		elif continentShape == "crab":
			shapeData = tbsCrabLike
	
	#calculate actual land tiles we must produce
	maxLandTiles = int(iW * (iH - continentSeparation * 2) * ((100 - oceanPercentage) / 100.0) * 0.7)
	
	#world is two-dimensional array which wraps in X automagically, so you may not worry about edges
	world = worldArray([iW,iH],PlotTypes.PLOT_OCEAN)
	
	#now we must somehow distribute start positions for each continent, depending on number of players.
	#if there is not enough continents for players, we will make some random continent to grow more.
	#actual player distribution is handled by Firaxis code, which looks how big each continent is and places players acording to that.
	if players > continentCount and continentSize == "players":
		# we will make helper-array to randomly make some continents bigger
		contData = []
		# first of all, we make continent amount the same as plaer amount, and add to each continent 1 player:
		for p in range(players):
			contData.append(1) # 1 player in 1 continent
		
		#colapse continents to match what we have
		#colapsing: if we in helper-array have more continents than we need, remove one item from array
		#  and randomly increase other array item by 1
		#do so until we get right continent count
		while len(contData) > continentCount:
			index = int(getRand(dice) * (len(contData) - 1)) #random index
			#merge it with nearest continent
			if index == 0:
				index2 = len(contData) - 1
			else:
				index2 = index - 1
			contData[index2] += contData[index]
			contData.pop(index)


		#position calculation moved to other function
		contPositions = getStartPositions(world, contData)

		#start continents
		contList = [] #contList will contain TileBuilders for each continent

		for i in range(len(contData)): #now contData contains how many players are in each continent
			if shapeRandom: #if we have random shape for each continent, we get it now
				index = int(getRand(dice) * (len(randomList) - 1)) #random index
				shapeData = randomList[index]
			#initialize tile builder
			continent = TileBuilder(world, PlotTypes.PLOT_LAND, shapeData[0], shapeData[1], shapeData[2], continentSeparation, maxLandTiles / players * contData[i])
			#place first tile (x, y)
			startX, startY = contPositions[i]
			continent.start(startX, startY)
			#add tile builder to builder list
			contList.append(continent)
			
		#now what is left - growing :)
		size = 0 #at first size is 0
		grows = True #variable to check if everything still grows
		#if size < maxLandTiles and grows:
		while size < maxLandTiles and grows:
			for i, continent in enumerate(contList): #for each index, TileBuilder in contList
				if contData[i] == 1: #if one player is supposed to be there
					if continent.grow(): #if tile builder is able to grow and grows one tile
						size += 1 #increase size by one
					else:
						grows = False #otherwise stop this script
				else: #if more than one player is there, grow it for each of them :)
					cGrows = True
					for j in range(contData[i]): # grow continent more if there are more players in it
						if continent.grow():
							size += 1
						else:
							cGrows = False
					if not cGrows:
						grows = False
	elif continentSize == "unpredictable":
		contPositions = getStartPositions(world, None)
		contList = []

		for i in range(continentCount):
			if shapeRandom:
				index = int(getRand(dice) * (len(randomList) - 1)) #random index
				shapeData = randomList[index]
			continent = TileBuilder(world, PlotTypes.PLOT_LAND, shapeData[0], shapeData[1], shapeData[2], continentSeparation, maxLandTiles / continentCount)
			startX, startY = contPositions[i]
			continent.start(startX, startY)
			contList.append(continent)
		
		#grow continents
		size = 0
		grows = True
		while size < maxLandTiles and grows:
		#if size < maxLandTiles and grows:
			for continent in contList:
				while getRand(dice) > 0.6:
					if continent.grow() and size < maxLandTiles:
						size += 1
					else:
						grows = False
						break
	else: #if there are less players than continents, code looks like English ;)
		#start continents
		contPositions = getStartPositions(world, None)
		contList = []

		for i in range(continentCount):
			if shapeRandom:
				index = int(getRand(dice) * (len(randomList) - 1)) #random index
				shapeData = randomList[index]
			continent = TileBuilder(world, PlotTypes.PLOT_LAND, shapeData[0], shapeData[1], shapeData[2], continentSeparation, maxLandTiles / continentCount)
			startX, startY = contPositions[i]
			continent.start(startX, startY)
			contList.append(continent)
		
		#grow continents
		size = 0
		grows = True
		while size < maxLandTiles and grows:
		#if size < maxLandTiles and grows:
			for continent in contList:
				if continent.grow():
					size += 1
				else:
					grows = False

	print "size",size
	
	# now we want to remove one-square and two-square lakes. more templates (which describe how to remove tiles) can be added in initReplacements function,
	#  and code below should handle them just fine. i am not adding comments to it because i forgot how it works :D
	rep = initReplacements()
	
	for x in range(iW):
		for y in range(iH):
			for r in rep:
				orig, final = r
				h = len(orig)
				w = len(orig[0])
				dh = (h - 1) / 2
				dw = (w - 1) / 2
				Tinka = True
				for ax in range(w):
					for ay in range(h):
						cell = orig[ay][ax]
						if landR(cell) != None:
							if landR(cell) != world[x-dw+ax,y-dh+ay]:
								Tinka = False
								break
				if Tinka:
					h = len(final)
					w = len(final[0])
					dh = (h - 1) / 2
					dw = (w - 1) / 2
					for ax in range(w):
						for ay in range(h):
							cell = final[ay][ax]
							if landR(cell) != None:
								world[x-dw+ax,y-dh+ay] = landR(cell)

	# now prepare to place peaks
	peakOptions = getSelValue("peaks")
	if peakOptions is None:
		peakOptions = [0,0,0]
	
	peakStartProbability = peakOptions[0]
	peakGrowProbability = peakOptions[1]
	maxPeaksInRow = peakOptions[2]

	if peakStartProbability > 0: #if we want peeks
		peakArray = bArray()

		for x in range(iW):
			for y in range(iH):
				#check if land area is 99 empty
				if world[x,y] == PlotTypes.PLOT_LAND and world[x-1,y] == PlotTypes.PLOT_LAND and world[x+1,y] == PlotTypes.PLOT_LAND and world[x,y-1] == PlotTypes.PLOT_LAND and world[x,y+1] == PlotTypes.PLOT_LAND and (not world[x+1,y+1] == PlotTypes.PLOT_PEAK) and (not world[x-1,y+1] == PlotTypes.PLOT_PEAK) and (not world[x-1,y-1] == PlotTypes.PLOT_PEAK) and (not world[x+1,y-1] == PlotTypes.PLOT_PEAK):
					if getRand(dice) < peakStartProbability:
						offset = world.getOffset([x,y])
						world._data[offset] = PlotTypes.PLOT_PEAK
						if peakGrowProbability > 0: #if we want peeks to grow
							peakArray[offset] = [x,y,1]
		
		# now grow peaks. this code is actually using simplified method of tileBuilder. BTW, i tried to use tileBuilder to grow
		#  peaks and hills and that worked fine, but in this way it is much more optimized and without unecessary things
		while len(peakArray) > 0:
			x, y, count = peakArray.random(dice) #get peak and remove it from array
			if count < maxPeaksInRow: # if we are in limit
				offset1 = world.getOffset([x,y])

				#print "++ Grow peak ++"
				#print "offset: %d"%(offset1)
				
				# for every peak we find all possible branches
				possibleTiles = bArray() # store them here
				c = 0
				wtt = world[x-1,y]
				if not wtt == PlotTypes.PLOT_PEAK and not peakIntersect(world,x-1,y,x,y):
					possibleTiles[world.getOffset([x-1,y])] = [x-1,y]
					c += 1
				wtt = world[x-1,y-1]
				if not wtt == PlotTypes.PLOT_PEAK and not peakIntersect(world,x-1,y-1,x,y):
					possibleTiles[world.getOffset([x-1,y-1])] = [x-1,y-1]
					c += 1
				wtt = world[x,y-1]
				if not wtt == PlotTypes.PLOT_PEAK and not peakIntersect(world,x,y-1,x,y):
					possibleTiles[world.getOffset([x,y-1])] = [x,y-1]
					c += 1
				wtt = world[x+1,y-1]
				if not wtt == PlotTypes.PLOT_PEAK and not peakIntersect(world,x+1,y-1,x,y):
					possibleTiles[world.getOffset([x+1,y-1])] = [x+1,y-1]
					c += 1
				wtt = world[x+1,y]
				if not wtt == PlotTypes.PLOT_PEAK and not peakIntersect(world,x+1,y,x,y):
					possibleTiles[world.getOffset([x+1,y])] = [x+1,y]
					c += 1
				wtt = world[x+1,y+1]
				if not wtt == PlotTypes.PLOT_PEAK and not peakIntersect(world,x+1,y+1,x,y):
					possibleTiles[world.getOffset([x+1,y+1])] = [x+1,y+1]
					c += 1
				wtt = world[x,y+1]
				if not wtt == PlotTypes.PLOT_PEAK and not peakIntersect(world,x,y+1,x,y):
					possibleTiles[world.getOffset([x,y+1])] = [x,y+1]
					c += 1
				wtt = world[x-1,y+1]
				if not wtt == PlotTypes.PLOT_PEAK and not peakIntersect(world,x-1,y+1,x,y):
					possibleTiles[world.getOffset([x-1,y+1])] = [x-1,y+1]
					c += 1
				
				tile = possibleTiles.random(dice) #take one tile from possible tiles
				
				if tile is not None and getRand(dice) < peakGrowProbability:
					ax, ay = tile
					world[ax,ay] = PlotTypes.PLOT_PEAK # add it
					peakArray[world.getOffset([ax,ay])] = [ax,ay,count + 1] # add it to peak array to make it continue group grow

	#now grow hills, the same thing as peaks, only this time we check for peaks when growing which obviosly already exist :)
	hillOptions = getSelValue("hills")
	if hillOptions is None:
		hillOptions = [0,0,0]
	
	hillStartProbability = hillOptions[0]
	hillGrowProbability = hillOptions[1]
	maxHillsInRow = hillOptions[2]

	if hillStartProbability > 0: #if we want peeks
		hillArray = bArray()

		for x in range(iW):
			for y in range(iH):
				#check if land area is 1111 empty
				if world[x,y] == PlotTypes.PLOT_LAND and not hillIntersect(world,x,y,x,y,1):
					if getRand(dice) < hillStartProbability:
						offset = world.getOffset([x,y])
						world._data[offset] = PlotTypes.PLOT_HILLS
						if hillGrowProbability > 0: #if we want peeks to grow
							hillArray[offset] = [x,y,1]

		while len(hillArray) > 0:
			x, y, count = hillArray.random(dice) #get peak and remove it from array
			if count < maxHillsInRow: # if we are in limit
				offset1 = world.getOffset([x,y])

				# for every hill we find all possible branches
				possibleTiles = bArray() # store them here
				c = 0
				separation = 1
				wtt = world[x-1,y]
				if wtt == PlotTypes.PLOT_LAND and not hillIntersect(world,x-1,y,x,y,separation):
					possibleTiles[world.getOffset([x-1,y])] = [x-1,y]
					c += 1
				wtt = world[x-1,y-1]
				if wtt == PlotTypes.PLOT_LAND and not hillIntersect(world,x-1,y-1,x,y,separation):
					possibleTiles[world.getOffset([x-1,y-1])] = [x-1,y-1]
					c += 1
				wtt = world[x,y-1]
				if wtt == PlotTypes.PLOT_LAND and not hillIntersect(world,x,y-1,x,y,separation):
					possibleTiles[world.getOffset([x,y-1])] = [x,y-1]
					c += 1
				wtt = world[x+1,y-1]
				if wtt == PlotTypes.PLOT_LAND and not hillIntersect(world,x+1,y-1,x,y,separation):
					possibleTiles[world.getOffset([x+1,y-1])] = [x+1,y-1]
					c += 1
				wtt = world[x+1,y]
				if wtt == PlotTypes.PLOT_LAND and not hillIntersect(world,x+1,y,x,y,separation):
					possibleTiles[world.getOffset([x+1,y])] = [x+1,y]
					c += 1
				wtt = world[x+1,y+1]
				if wtt == PlotTypes.PLOT_LAND and not hillIntersect(world,x+1,y+1,x,y,separation):
					possibleTiles[world.getOffset([x+1,y+1])] = [x+1,y+1]
					c += 1
				wtt = world[x,y+1]
				if wtt == PlotTypes.PLOT_LAND and not hillIntersect(world,x,y+1,x,y,separation):
					possibleTiles[world.getOffset([x,y+1])] = [x,y+1]
					c += 1
				wtt = world[x-1,y+1]
				if wtt == PlotTypes.PLOT_LAND and not hillIntersect(world,x-1,y+1,x,y,separation):
					possibleTiles[world.getOffset([x-1,y+1])] = [x-1,y+1]
					c += 1
				
				tile = possibleTiles.random(dice) #take one tile from possible tiles
				
				if tile is not None and getRand(dice) < hillGrowProbability:
					ax, ay = tile
					world[ax,ay] = PlotTypes.PLOT_HILLS # add it
					hillArray[world.getOffset([ax,ay])] = [ax,ay,count + 1] # add it to peak array to make it continue group grow

	# return our world in one-dimensional array to Firaxis to do the rest :)
	return world.get()

def getStartPositions(world, contData):
	global continentCount
	global continentSeparation

	cgc = CyGlobalContext()
	map = cgc.getMap()

	distribution = getContinentDistribution()

	dice = cgc.getGame().getMapRand() #dice is Firaxis' random number generator
	players = cgc.getGame().countCivPlayersEverAlive()
	iW = map.getGridWidth()
	iH = map.getGridHeight()

	if contData == None:
		contData = []
		for p in range(continentCount):
			contData.append(1) # create dummy array
	
	positionArray = []

	if distribution == "equator": #old style distribution
		p = -1
		part = float(iW) / float(continentCount * 2) #figure out x position of continent start tile
		for i in range(len(contData)):
			p += 2
			positionArray.append([int(part * p), iH / 2])
	elif distribution == "2rows":
		p = -1
		direction = 1
		part = float(iW) / float(continentCount * 2) #figure out x position of continent start tile
		for i in range(len(contData)):
			p += 2
			positionArray.append([int(part * p), int(iH / 2 * (1 + direction * (0.2 + getRand(dice) * 0.3)))])
			direction *= -1
	elif distribution == "3rows":
		p = -1
		offset = 0
		offsetchange = 1
		part = float(iW) / float(continentCount * 2) #figure out x position of continent start tile
		for i in range(len(contData)):
			p += 2
			positionArray.append([int(part * p), int(iH / 2 * (1 + offset * (0.4 + getRand(dice) * 0.2)))])
			offset += offsetchange
			if offset > 0 or offset < 0:
				offsetchange *= -1
	elif distribution == "center":
		perimeter = (continentSeparation + 3) * continentCount
		radius = perimeter / (2 * 3.141592)
		inc_angle = (360 / continentCount) * 3.1415 / 180
		angle = 0
		height_multiplier = 0.9
		width_multiplier = 1.5
		for i in range(continentCount):
			if angle >= 360 * 3.1415 / 180:
				angle -= 360 * 3.1415 / 180
			if angle == 0:
				positionArray.append([ int(iW / 2 - radius * width_multiplier), int(iH / 2) ])
			elif angle > 0 and angle < 3.1415 / 2:
				positionArray.append([ int(iW / 2 - radius * width_multiplier * math.cos(angle)), int(iH / 2 - radius * height_multiplier * math.sin(angle)) ])
			elif angle >= 3.1415 / 2 and angle < 3.1415:
				positionArray.append([ int(iW / 2 + radius * width_multiplier * math.sin(angle - 3.1415 / 2)), int(iH / 2 - radius * height_multiplier * math.cos(angle - 3.1415 / 2)) ])
			elif angle >= 3.1415 and angle < 3.1415 / 2 * 3:
				positionArray.append([ int(iW / 2 + radius * width_multiplier * math.cos(angle - 3.1415)), int(iH / 2 + radius * height_multiplier * math.sin(angle - 3.1415)) ])
			else:
				positionArray.append([ int(iW / 2 - radius * width_multiplier * math.sin(angle - 3.1415 / 2 * 3)), int(iH / 2 + radius * height_multiplier * math.cos(angle - 3.1415 / 2 * 3)) ])
				
			angle += inc_angle

	return positionArray

def getContinentDistribution():
	global continentCount
	global continentSeparation

	distributionOption = getSelValue("continent_distribution")
	if distributionOption == None:
		distributionOption = "smart-random"

	cgc = CyGlobalContext()
	map = cgc.getMap()

	worldSize = map.getWorldSize()

	sizeDuel = cgc.getInfoTypeForString("WORLDSIZE_DUEL")
	sizeTiny = cgc.getInfoTypeForString("WORLDSIZE_TINY")
	sizeSmall = cgc.getInfoTypeForString("WORLDSIZE_SMALL")
	sizeStandart = cgc.getInfoTypeForString("WORLDSIZE_STANDARD")
	sizeLarge = cgc.getInfoTypeForString("WORLDSIZE_LARGE")
	sizeHuge = cgc.getInfoTypeForString("WORLDSIZE_HUGE")

	if distributionOption == "smart-random":
		if worldSize == sizeDuel:
			if continentSeparation < 4:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 15:
					distribution = "2rows"
				else:
					distribution = "3rows"
			else:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 12:
					distribution = "2rows"
				else:
					distribution = "3rows"
		elif worldSize == sizeTiny:
			if continentSeparation < 4:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 13:
					distribution = "2rows"
				else:
					distribution = "3rows"
			else:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 12:
					distribution = "2rows"
				else:
					distribution = "3rows"
		elif worldSize == sizeSmall:
			if continentSeparation < 4:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 14:
					distribution = "2rows"
				else:
					distribution = "3rows"
			else:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 13:
					distribution = "2rows"
				else:
					distribution = "3rows"
		elif worldSize == sizeStandart:
			if continentSeparation < 4:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 15:
					distribution = "2rows"
				else:
					distribution = "3rows"
			else:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 14:
					distribution = "2rows"
				else:
					distribution = "3rows"
		elif worldSize == sizeLarge:
			if continentSeparation < 4:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 16:
					distribution = "2rows"
				else:
					distribution = "3rows"
			else:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 15:
					distribution = "2rows"
				else:
					distribution = "3rows"
		elif worldSize == sizeHuge:
			if continentSeparation < 4:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 16:
					distribution = "2rows"
				else:
					distribution = "3rows"
			else:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 15:
					distribution = "2rows"
				else:
					distribution = "3rows"
		else:
			if continentSeparation < 4:
				if continentCount < 3:
					distribution = "center"
				elif continentCount < 17:
					distribution = "2rows"
				else:
					distribution = "3rows"
			else:
				if continentCount < 4:
					distribution = "center"
				elif continentCount < 15:
					distribution = "2rows"
				else:
					distribution = "3rows"
	else:
		distribution = distributionOption
	
	return distribution

#function to determine if tile x,y intersects with something, ignoring ox,oy tile
def peakIntersect(world, x, y, ox, oy):
	ax = x
	ay = y + 1
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world._data[offset] == PlotTypes.PLOT_PEAK or world._data[offset] == PlotTypes.PLOT_OCEAN:
			return True
	ax = x + 1
	ay = y
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world._data[offset] == PlotTypes.PLOT_PEAK or world._data[offset] == PlotTypes.PLOT_OCEAN:
			return True
	ax = x
	ay = y - 1
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world._data[offset] == PlotTypes.PLOT_PEAK or world._data[offset] == PlotTypes.PLOT_OCEAN:
			return True
	ax = x - 1
	ay = y
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world._data[offset] == PlotTypes.PLOT_PEAK or world._data[offset] == PlotTypes.PLOT_OCEAN:
			return True

	ax = x - 1
	ay = y - 1
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world._data[offset] == PlotTypes.PLOT_PEAK:
			return True
	ax = x - 1
	ay = y + 1
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world._data[offset] == PlotTypes.PLOT_PEAK:
			return True
	ax = x + 1
	ay = y + 1
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world._data[offset] == PlotTypes.PLOT_PEAK:
			return True
	ax = x + 1
	ay = y - 1
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world._data[offset] == PlotTypes.PLOT_PEAK:
			return True

	return False

#function to determine if tile x,y intersects with something in distance <separation>, ignoring ox,oy tile
def hillIntersect(world, x, y, ox, oy, separation):
	ax = x - 1
	ay = y
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world[ax, ay] == PlotTypes.PLOT_HILLS:
			return True
	ax = x + 1
	ay = y
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world[ax, ay] == PlotTypes.PLOT_HILLS:
			return True
	ax = x
	ay = y - 1
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world[ax, ay] == PlotTypes.PLOT_HILLS:
			return True
	ax = x
	ay = y + 1
	if ax != ox and ay != oy:
		offset = world.getOffset([ax, ay])
		if world[ax, ay] == PlotTypes.PLOT_HILLS:
			return True
	return False

#------------------------------------------------------
"""GAME	ELEMENT	PLACEMENT FUNCTIONS"""
#------------------------------------------------------

def	beforeInit():
	"Called	before the map is initialized"

def	beforeGeneration():
	"Called	before the map generation starts"

# default implementation
def	addRivers():
	"Should	add	all	desired	rivers to the map."
	CyPythonMgr().allowDefaultImpl()

# default implementation
def	getRiverStartCardinalDirection(argsList):
	pPlot =	argsList[0]
	"Returns the cardinal direction	of the first river segment."
	CyPythonMgr().allowDefaultImpl()

# default implementation
def	getRiverAltitude(argsList):
	pPlot =	argsList[0]
	"Returns the altitude level	of pPlot (rivers will flow to the lowest point)."
	CyPythonMgr().allowDefaultImpl()

# default implementation
def	addLakes():
	"Can override to control where lakes are added on the map"
	CyPythonMgr().allowDefaultImpl()
	
#Step 4: add features to your terrain	
#this is responsible for things like oasis, ice floes on the ocean, floodplains
def	addFeatures():
	global world

	global climateVariation
	global climateTemperature

	global climateHumidity
	climateHumidity = getSelValue("humidity")
	if climateHumidity is None:
		climateHumidity = 0.5
	floodPlains = getSelValue("flood_plains")
	if floodPlains is None:
		floodPlains = -1
	
	iceOnWater = 0.5

	global continentSeparation

	cgc = CyGlobalContext()

	terrainDesert = cgc.getInfoTypeForString("TERRAIN_DESERT")
	terrainPlains = cgc.getInfoTypeForString("TERRAIN_PLAINS")
	terrainGrass = cgc.getInfoTypeForString("TERRAIN_GRASS")
	terrainTundra = cgc.getInfoTypeForString("TERRAIN_TUNDRA")
	terrainSnow = cgc.getInfoTypeForString("TERRAIN_SNOW")

	featNone = cgc.getInfoTypeForString("NO_FEATURE")
	featIce = cgc.getInfoTypeForString("FEATURE_ICE")
	featJungle = cgc.getInfoTypeForString("FEATURE_JUNGLE")
	featOasis = cgc.getInfoTypeForString("FEATURE_OASIS")
	featFlood = cgc.getInfoTypeForString("FEATURE_FLOOD_PLAINS")
	featForest = cgc.getInfoTypeForString("FEATURE_FOREST")
	
	map = cgc.getMap()
	dice = cgc.getGame().getMapRand()
	players = cgc.getGame().countCivPlayersEverAlive()
	iW = map.getGridWidth()
	iH = map.getGridHeight()

	# Terrain Type generation
	seedList = bArray()
	for y in range(iH):
		randomTerrain = probabilityArray()

		tileTemperature = getTileTemperature(y,iH)
		print "tile temperature",tileTemperature
		if tileTemperature > 30:
			randomTerrain[7 * (1.5 - climateHumidity)] = terrainDesert
		if tileTemperature > 15 and tileTemperature < 39:
			randomTerrain[7] = terrainPlains
		if tileTemperature > 4 and tileTemperature < 30:
			randomTerrain[7 * (climateHumidity + 0.5)] = terrainGrass
		if tileTemperature > -2 and tileTemperature < 25:
			randomTerrain[7] = terrainPlains
		if tileTemperature > -10 and tileTemperature < 10:
			randomTerrain[7] = terrainTundra
		if tileTemperature < -30:
			randomTerrain[15 * (climateHumidity / 2 + 0.75)] = terrainSnow
		elif tileTemperature < 0:
			randomTerrain[7 * (climateHumidity / 2 + 0.75)] = terrainSnow

		for x in range(iW):
			plot = map.sPlot(x,y)
			if plot.isWater():
				if y < continentSeparation and not plot.isAdjacentToLand():
					prob = (continentSeparation - y) / float(continentSeparation)
					if getRand(dice) < prob:
						plot.setFeatureType(featIce,-1)
				elif y > iH - continentSeparation - 1 and not plot.isAdjacentToLand():
					prob = (continentSeparation - (iH - y - 1)) / float(continentSeparation)
					if getRand(dice) < prob:
						plot.setFeatureType(featIce,-1)

				if plot.isAdjacentToLand() and y > continentSeparation and y < iH - continentSeparation - 1:
					if tileTemperature < -40:
						if getRand(dice) < iceOnWater * 2:
							plot.setFeatureType(featIce,-1)
					elif tileTemperature < -25:
						if getRand(dice) < iceOnWater:
							plot.setFeatureType(featIce,-1)
					elif tileTemperature < -10:
						if getRand(dice) < iceOnWater / 2:
							plot.setFeatureType(featIce,-1)
					elif tileTemperature < -5:
						if getRand(dice) < iceOnWater / 3:
							plot.setFeatureType(featIce,-1)
					elif tileTemperature < 0:
						if getRand(dice) < iceOnWater / 4:
							plot.setFeatureType(featIce,-1)
			else:
				if not plot.isPeak() and (plot.isFreshWater() or plot.isCoastalLand()):
					seedList[world.getOffset([x,y])] = [x,y,1,None] # x,y,distance from river,FeatureType
				plot.setTerrainType(randomTerrain.randomItem(dice),False,False)
	
	# Feature generation
	while len(seedList) > 0:
		x, y, distanceFromRiver, feature = seedList.random(dice)
		plot = map.sPlot(x,y)
		terrain = plot.getTerrainType()
		tileTemperature = getTileTemperature(y,iH)
		
		if feature == None and plot.getFeatureType() == featNone:
			randomFeat = probabilityArray()
			
			randomFeat[distanceFromRiver * 20 * (1 - climateHumidity)] = None # how many tiles have no forest and no jungle
			
			# Features based on initial terrain
			if terrain == terrainDesert:
				randomFeat[1] = featJungle #more jungle on sand
			elif terrain == terrainGrass:
				if tileTemperature > 20:
					randomFeat[1] = featJungle
				randomFeat[3] = featForest #most forest on grass
			elif terrain == terrainPlains:
				if tileTemperature > 25:
					randomFeat[2] = featJungle
				randomFeat[2] = featForest
			elif terrain == terrainTundra:
				randomFeat[2] = featForest #more forest in tundra
			elif terrain == terrainSnow:
				randomFeat[2] = featForest #some forest on snow
			if terrain != terrainTundra and terrain != terrainSnow: #no jungle on these
				if tileTemperature > 30:
					randomFeat[1] = featJungle
					if climateHumidity > 0.8:
						randomFeat[1] = featJungle
			
			# Floods:
			if floodPlains < 0: #standard generation
				if (not plot.isHills()) and plot.isFreshWater(): #no floods on hills
					if getRand(dice) > -1 * floodPlains:
						if terrain == terrainDesert:
							randomFeat[8 * (1 - climateHumidity)] = featFlood
						elif terrain == terrainTundra:
							randomFeat[6 * (1 - climateHumidity)] = featFlood			
			else:
				if not plot.isHills(): #no floods on hills
					randomFeat[floodPlains] = featFlood
					if plot.isFreshWater(): #make more floods near lakes and rivers
						if terrain == terrainTundra or terrain == terrainSnow:
							if tileTemperature > -10:
								randomFeat[2 * (1 - climateHumidity)] = featFlood
							elif tileTemperature > -30:
								randomFeat[1 * (1 - climateHumidity)] = featFlood
						elif terrain == terrainDesert:
							if tileTemperature < 40:
								randomFeat[2 * (1 - climateHumidity)] = featFlood
							elif tileTemperature < 55:
								randomFeat[1 * (1 - climateHumidity)] = featFlood
						else:
							randomFeat[4 * (1 - climateHumidity)] = featFlood
					# We increase amount of flood on poor terrain, to make it more playable
					if terrain == terrainDesert:
						randomFeat[3 * (1 - climateHumidity)] = featFlood
					elif terrain == terrainSnow:
						randomFeat[3 * (1 - climateHumidity)] = featFlood
					elif terrain == terrainTundra:
						randomFeat[1 * (1 - climateHumidity)] = featFlood

			# Features based on temperature
			if tileTemperature > 40:
				randomFeat[2] = featJungle
			elif tileTemperature > 15:
				randomFeat[2] = featForest
			elif tileTemperature > 0:
				randomFeat[1] = featForest
			elif tileTemperature > -20:
				randomFeat[2] = featForest
			else:
				randomFeat[1] = featForest
			
			feat = randomFeat.randomItem(dice)
			if feat != None:
				plot.setFeatureType(feat,-1)
				feature = feat

		if feature != None:
			if plot.getFeatureType() == featNone:
				plot.setFeatureType(feature,-1)

			#feature, when spreads, modifies terrain type
			if feature == featJungle:
				if plot.isFreshWater():
					if getRand(dice) > 0.4:
						plot.setTerrainType(terrainPlains,False,False)
					else:
						plot.setTerrainType(terrainGrass,False,False)
				elif terrain == terrainDesert:
					if getRand(dice) > 0.05:
						if getRand(dice) > 0.3:
							plot.setTerrainType(terrainPlains,False,False)
						else:
							plot.setTerrainType(terrainGrass,False,False)
				elif terrain == terrainPlains:
					if getRand(dice) > 0.2:
						plot.setTerrainType(terrainGrass,False,False)
				elif terrain == terrainTundra:
					plot.setFeatureType(featForest,-1)
				elif terrain == terrainSnow: #that happened once, looks weird :)
					plot.setFeatureType(featForest,-1)
			elif feature == featForest:
				if terrain == terrainDesert:
					if getRand(dice) > 0.3:
						plot.setTerrainType(terrainPlains,False,False)
					else:
						plot.setTerrainType(terrainGrass,False,False)
				elif terrain == terrainPlains:
					if getRand(dice) > 0.4:
						plot.setTerrainType(terrainGrass,False,False)
			elif feature == featFlood:
				if (not plot.isFreshWater()) and floodPlains >= 0:
					if terrain == terrainDesert:
						if getRand(dice) > 0.8:
							plot.setTerrainType(terrainPlains,False,False)
					elif terrain == terrainPlains:
						if getRand(dice) > 0.8:
							plot.setTerrainType(terrainGrass,False,False)
					elif terrain == terrainSnow:
						if getRand(dice) > 0.9:
							plot.setTerrainType(terrainTundra,False,False)
					
			# make this seed spread to other tiles
			tileList = getTilesAround(x,y)
			posList = probabilityArray()
			for tile in tileList:
				tx,ty = tile
				tilePlot = map.sPlot(tx,ty)
				tileTerrain = tilePlot.getTerrainType()
				if not tilePlot.isWater() and not tilePlot.isPeak():
					if feature == featFlood:
						if floodPlains > 0: #if this is not standart mode
							if tilePlot.getFeatureType() == featNone and tilePlot.isFlatlands():
								if tilePlot.isFreshWater():
									posList[2] = [tx,ty,distanceFromRiver + 3,feature]
								else:
									posList[1] = [tx,ty,distanceFromRiver + 3,feature]
					elif feature == featJungle:
						if tilePlot.getFeatureType() == featNone:
							if tilePlot.isFlatlands():
								posList[1] = [tx,ty,distanceFromRiver + 1,feature]
							if not isRiverCrossing(tilePlot,plot):
								if tileTerrain != terrainSnow and tileTerrain != terrainTundra:
									if tileTerrain == terrainPlains:
										posList[1] = [tx,ty,distanceFromRiver + 1,feature]
									else:
										posList[2] = [tx,ty,distanceFromRiver + 1,feature]
					elif feature == featForest:
						if tilePlot.getFeatureType() == featNone:
							if tilePlot.isFlatlands():
								posList[1] = [tx,ty,distanceFromRiver + 1,feature]
							if not isRiverCrossing(tilePlot,plot):
								if tileTerrain != terrainDesert:
									if tileTerrain == terrainPlains:
										posList[1] = [tx,ty,distanceFromRiver + 1,feature]
									else:
										posList[2] = [tx,ty,distanceFromRiver + 1,feature]
			
			posList[len(posList) / 2] = None # add some none probability
			item = posList.random(dice)
			if item != None:
				seedList[world.getOffset([item[0],item[1]])] = item

	# fix terrain near jungle
	for y in range(iH):
		for x in range(iW):
			plot = map.sPlot(x,y)
			if plot.getFeatureType() == featJungle:
				tileList = getTilesAround(x,y)
				for tile in tileList:
					tx,ty = tile
					tilePlot = map.sPlot(tx,ty)
					tileTerrain = tilePlot.getTerrainType()
					if tileTerrain == terrainDesert:
						if (getRand(dice) > 0.9):
							plot.setFeatureType(featNone,-1)
							break
						else:
							if tilePlot.isFreshWater():
								if (tilePlot.getFeatureType() == featFlood):
									if (getRand(dice) > 0.5):
										tilePlot.setTerrainType(terrainPlains,False,False)
								else:
									if climateHumidity > 0.5:
										if (getRand(dice) > 0.4):
											tilePlot.setTerrainType(terrainPlains,False,False)
										else:
											tilePlot.setTerrainType(plot.getTerrainType(),False,False)
									else:
										if (getRand(dice) > 0.2):
											tilePlot.setTerrainType(terrainPlains,False,False)
										else:
											tilePlot.setTerrainType(plot.getTerrainType(),False,False)
							else:
								if (tilePlot.getFeatureType() == featFlood):
									if (getRand(dice) > 0.8):
										tilePlot.setTerrainType(terrainPlains,False,False)
								else:
									if climateHumidity > 0.5:
										if (getRand(dice) > 0.3):
											tilePlot.setTerrainType(terrainPlains,False,False)
										else:
											tilePlot.setTerrainType(plot.getTerrainType(),False,False)
									else:
										if (getRand(dice) > 0.1):
											tilePlot.setTerrainType(terrainPlains,False,False)

					elif tileTerrain == terrainTundra:
						if plot.getTerrainType() == terrainGrass:
							tilePlot.setTerrainType(terrainPlains,False,False)
						else:
							tilePlot.setTerrainType(plot.getTerrainType(),False,False)
					elif tileTerrain == terrainSnow:
						if (getRand(dice) > 0.5):
							tilePlot.setTerrainType(terrainPlains,False,False)
						else:
							tilePlot.setTerrainType(terrainTundra,False,False)

	# find best places for oasis
	possibleOasisList = bArray()
	for y in range(iH):
		for x in range(iW):
			plot = map.sPlot(x,y)
			if (not plot.isWater()) and (not plot.isFreshWater()) and (not plot.isHills()) and (not plot.isCoastalLand()) and (not plot.isPeak()):
				score = 0
				if plot.getTerrainType() == terrainDesert:
					score += 10
					tileList = getTilesAround(x,y)
					for tile in tileList:
						tx,ty = tile
						tilePlot = map.sPlot(tx,ty)
						tileTerrain = tilePlot.getTerrainType()
						if tilePlot.isFreshWater():
							score -= 20
						if tilePlot.isRiverSide():
							score -= 20
						if tilePlot.isPeak():
							score -= 2
						if tilePlot.getFeatureType() == featNone:
							score += 1
						if tileTerrain == terrainDesert:
							score += 1
						elif tileTerrain == terrainPlains:
							score -= 5
						elif tileTerrain == terrainGrass:
							score -= 20
						elif tileTerrain == terrainTundra:
							score -= 20
						elif tileTerrain == terrainSnow:
							score -= 20
						if not tilePlot.hasYield():
							score += 1

				if score > 0:
					#plot.setFeatureType(featOasis,-1)
					possibleOasisList[world.getOffset([x,y])] = [x, y]
	
	missedItems = 0
	for index in range(len(possibleOasisList) / 4):
		x, y = possibleOasisList.random(dice)
		canPlace = True
		tileList = getTilesAround(x,y)
		for tile in tileList:
			tx,ty = tile
			plot = map.sPlot(tx,ty)
			if plot.getFeatureType() == featOasis:
				canPlace = False
				missedItems += 1
				break
		plot = map.sPlot(x,y)
		if canPlace:
			plot.setFeatureType(featOasis,-1)
		if missedItems > 20:
			break

def isRiverCrossing(plot1,plot2):
	cgc = CyGlobalContext()

	dirNorth = cgc.getInfoTypeForString("DIRECTION_NORTH")
	dirNorthEast = cgc.getInfoTypeForString("DIRECTION_NORTHEAST")
	dirEast = cgc.getInfoTypeForString("DIRECTION_EAST")
	dirSouthEast = cgc.getInfoTypeForString("DIRECTION_SOUTHEAST")
	dirSouth = cgc.getInfoTypeForString("DIRECTION_SOUTH")
	dirSouthWest = cgc.getInfoTypeForString("DIRECTION_SOUTHWEST")
	dirWest = cgc.getInfoTypeForString("DIRECTION_WEST")
	dirNorthWest = cgc.getInfoTypeForString("DIRECTION_NORTHWEST")

	map = cgc.getMap()
	iW = map.getGridWidth()
	iH = map.getGridHeight()

	# quite nasty lines of code to determine plot direction

	x1 = plot1.getX()
	y1 = plot1.getY()
	x2 = plot2.getX()
	y2 = plot2.getY()

	if x2 >= iW:
		x2 = 0
	elif x2 < 0:
		x2 = iW - 1
	if y2 >= iH:
		y2 = 0
	elif y2 < 0:
		y2 = iH - 1

	dx = x2 - x1
	dy = y2 - y1

	direction = -1
	if dx < 0 and dy == 0:
		direction = dirWest
	elif dx < 0 and dy > 0:
		direction = dirNorthWest
	elif dx == 0 and dy > 0:
		direction = dirNorth
	elif dx > 0 and dy > 0:
		direction = dirNorthEast
	elif dx > 0 and dy == 0:
		direction = dirEast
	elif dx > 0 and dy < 0:
		direction = dirSouthEast
	elif dx == 0 and dy < 0:
		direction = dirSouth
	elif dx < 0 and dy < 0:
		direction = dirSouthWest

	if direction == -1:
		return False
	else:
		return plot1.isRiverCrossing(direction)
	
def getTilesAround(x,y):
	cgc = CyGlobalContext()
	map = cgc.getMap()
	iW = map.getGridWidth()
	iH = map.getGridHeight()
	l = [[x-1,y],[x-1,y+1],[x,y+1],[x+1,y+1],[x+1,y],[x+1,y-1],[x,y-1],[x-1,y-1]]
	rem = []
	for index, item in enumerate(l):
		ax, ay = item
		if ay < 0 or ay >= iH:
			rem.append(index)
		if ax < 0:
			ax = iW - 1
			l[index] = [ax,ay]
		elif ax >= iW:
			ax = 0
			l[index] = [ax,ay]
	for index in rem:
		l.pop(index)
	return l

def initClimateSettings():
	global climateVariation
	global climateTemperature
	climateVariation = getSelValue("variation")
	if climateVariation is None:
		climateVariation = 0.4
	climateTemperature = getSelValue("temperature")
	if climateTemperature is None:
		climateTemperature = 40

def getTileTemperature(y,h):
	global climateTemperature
	global climateVariation
	lowestTemperature = ((climateTemperature + 50) * climateVariation) - 50
	th = h - 1
	half = th / 2.0
	if y > half:
		return lowestTemperature + ((climateTemperature - lowestTemperature) * (half - (y - half)) / half)
	else:
		return lowestTemperature + ((climateTemperature - lowestTemperature) * y / half)

#Step 5: add bonuses to your map
#this method puts down the tile bonuses, such as iron, whale, rice, oil, etc.	
def	addBonuses():
	"Can override to control where bonuses are added on	the	map"
	CyPythonMgr().allowDefaultImpl()

#Step 5B: override addBonusType so that it can pass through
def	addBonusType(argsList):
	"Can override to control where a type of bonus is added	on the map"
	"""WARNING:	If addBonuses()	is overriden then this function	will not automatically be called."""
	[iBonusType] = argsList
	CyPythonMgr().allowDefaultImpl()

def	canPlaceBonusAt(argsList):
	"Can override to determine if a	bonus can be placed	at the specified plot"
	"""WARNING:	If addBonuses()	is overriden then this function	will not automatically be called."""
	pPlot =	argsList[0]
	CyPythonMgr().allowDefaultImpl()

#Step 6: place goody huts
def	addGoodies():
	"Can override to control where goodies are added on	the	map"
	global generationStarted
	generationStarted = False
	CyPythonMgr().allowDefaultImpl()

def	canPlaceGoodyAt(argsList):
	pPlot =	argsList[0]
	"Can override to determine if a	goody can be placed	at the specified plot"
	CyPythonMgr().allowDefaultImpl()

def	afterGeneration():
	"Called	after the map generation is	finished"

#------------------------------------------------------
"""STARTING	PLOT FUNCTIONS"""
#------------------------------------------------------

def	minStartingDistanceModifier():
	"Returns a percent modification	for	the	minimum	distance between players (return 50	would mean +50%)"
	return 0

def	assignStartingPlots():
	"Can assign	as many	starting locations as desired"
	CyPythonMgr().allowDefaultImpl()

def	findStartingArea(argsList):
	"""Can override	to return the ID of	the	area in	which the given	player should start. Return	an areaID to
	force the player to	start in a certain area, or	-1 to specify that the player should start in the best
	location, irrespective of which	area it	is in. If not overridden, the default implementation will try
	to maximize	the	size of	the	player's part of his area
	
	WARNING: If	findStartingPlot() is overriden	then this function will	not	automatically be called."""
	playerID = argsList[0]
	CyPythonMgr().allowDefaultImpl()

def	findStartingPlot(argsList):
	"Can override to return	the	plot index at which	the	given player should	start"
	playerID = argsList[0]
	CyPythonMgr().allowDefaultImpl()

def	normalizeStartingPlotLocations():
	"Can override to change	how	starting plots are distributed.	(note, this	function is	called after all starting locs are distributed...)"
	CyPythonMgr().allowDefaultImpl()

def	normalizeAddRiver():
	"A normalize start function	(called	after starting plots are set) -	this call adds a river."
	#CyPythonMgr().allowDefaultImpl()

def	normalizeRemovePeaks():
	"A normalize start function	(called	after starting plots are set) -	this call removes peaks."
	#CyPythonMgr().allowDefaultImpl()

def	normalizeAddLakes():
	"A normalize start function	(called	after starting plots are set) -	this call adds a lake."
	#CyPythonMgr().allowDefaultImpl()

def	normalizeRemoveBadFeatures():
	"A normalize start function	(called	after starting plots are set) -	this call removes bad features."
	CyPythonMgr().allowDefaultImpl()

def	normalizeRemoveBadTerrain():
	"A normalize start function	(called	after starting plots are set) -	this call removes bad terrain."
	#CyPythonMgr().allowDefaultImpl()

def	normalizeAddFoodBonuses():
	"A normalize start function	(called	after starting plots are set) -	this call adds food	bonuses."
	#CyPythonMgr().allowDefaultImpl()

def	normalizeAddGoodTerrain():
	"A normalize start function	(called	after starting plots are set) -	this call adds good	terrain."
	#CyPythonMgr().allowDefaultImpl()

def	normalizeAddExtras():
	"A normalize start function	(called	after starting plots are set) -	this call adds extra features and bonuses."
	CyPythonMgr().allowDefaultImpl()

def	startHumansOnSameTile():
	"Returns true if all human units should	start on the same tile"
	return False

#----------------------------------------------
"""Helpers"""
#----------------------------------------------
def getRand(dice):
	"""return random number from 0 to 1"""
	return (dice.get(1001,"Python - EarthGen: Random dice.")) / 1000.0
# taken from SmartMap
import os
import os.path

if (sys.platform == 'darwin'):
	def civFilePath():
		civ4Dir = os.path.basename(os.getcwd())
		if "Warlords" in civ4Dir:
			civ4Dir = "Civilization IV Warlords"
		myDocuments = os.path.join(os.environ['HOME'], "Documents")
		finalFolder =  os.path.join(myDocuments, civ4Dir)
		return finalFolder

else:
	import _winreg
	def regRead(registry, path, field):
		try:
			pathKey = _winreg.OpenKey(registry, path)
			fieldValue = _winreg.QueryValueEx(pathKey, field)
			return fieldValue[0]
		finally:
			pathKey.Close()

	def civFilePath():
		try:
			userFolder = regRead(_winreg.HKEY_CURRENT_USER,"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders","Personal")
			#print "    userFolder",userFolder
			civFolder = os.path.basename(regRead(_winreg.HKEY_LOCAL_MACHINE,"Software\\Firaxis Games\\Sid Meier's Civilization 4","INSTALLDIR"))
			#print "    civFolder",civFolder
			finalFolder = os.path.join(os.path.join(userFolder, "My Games"), civFolder)
			#print "    finalFolder",finalFolder
			return finalFolder
		except:
			return ""
#----------------------------------------------
"""Other functions"""
#----------------------------------------------
def isAdvancedMap():
	"""
	Advanced maps only show	up in the map script pulldown on the advanced menu.
	Return 0 if	you	want your map to show up in	the	simple singleplayer	menu
	"""
	return 0
def getModPath():
	"""
	Returns	the	path of	the	mod	this map is	associated with
	"""
	return ""
def isClimateMap():
	"""Uses the Climate options"""
	return 0

def isSeaLevelMap():
	"""Uses the Sea Level options"""
	return 0

def getNumCustomMapOptions():
	loadMapOptionDefaults()
	return len(selection_names_and_values)

def getCustomMapOptionName(argsList):
	print "getCustomMapOptionName",argsList
	index =	argsList[0]
	return unicode(selection_titles[index][0])

def getNumCustomMapOptionValues(argsList):
	print "getNumCustomMapOptionValues",argsList
	index =	argsList[0]
	return len(selection_names_and_values[index])

def getCustomMapOptionDescAt(argsList):
	print "getCustomMapOptionDescAt",argsList
	iOption = argsList[0]
	iSelection = argsList[1]
	translated_text = unicode(selection_names_and_values[iOption][iSelection][0])
	return translated_text

def getCustomMapOptionDefault(argsList):
	index =	argsList[0]
	return selection_defaults[index]

def saveMapOptionDefaults():
	cymap = CyMap()
	fileName = os.path.join(civFilePath(),"planet_generator_" + VERSION + ".cfg")
	try:
		settings = open(fileName, 'w')
		smoptions = []
		index = -1
		for item in selection_titles:
			smoption = item[1]
			index += 1
			selectionIndex = int(cymap.getCustomMapOption(index))
			valueName = selection_names_and_values[index][selectionIndex][0]
			smoptions.append((smoption,valueName))
			
		print "    try to store settings to",fileName
		print smoptions
		try:
			pickle.dump(smoptions, settings)
		except Exception, inst:
			print "    Pickling Error",inst,"trying to save map settings to",fileName
		settings.close()
		print "    store successful"
	except IOError:
		print "    Couldn't create ",fileName
	except EOFError:
		print "    EOF writing ",fileName
	except:
		print "    unexpected problem writing ",fileName

def loadMapOptionDefaults():
	global selection_defaults
	fileName = os.path.join(civFilePath(),"planet_generator_" + VERSION + ".cfg")
	try:
		settings = open(fileName, 'r')
		optionsWithDefault = pickle.load(settings)
		#print optionsWithDefault
		for optionIndex in range(len(optionsWithDefault)):
			option, valueName = optionsWithDefault[optionIndex]
			for selOptionIndex in range(len(selection_titles)):
				selectionTitleCaption, selectionTitleName, autoRandom = selection_titles[selOptionIndex]
				if option == selectionTitleName:
					for selectionIndex in range(len(selection_names_and_values[selOptionIndex])):
						selectionName, selectionValue, allowedRandomPick = selection_names_and_values[selOptionIndex][selectionIndex]
						if selectionName == valueName:
							selection_defaults[selOptionIndex] = selectionIndex
		#print "    loaded option from cfg:",option,value
		settings.close()
	except IOError:
		print "    Couldn't find ",fileName
	except EOFError:
		print "    Bad contents in ",fileName
	#except:
	#	print "    unexpected problem reading",fileName

def isRandomCustomMapOption(argsList):
	return False # we use our own defaults

#----------------------------------------------
"""MAP SIZE	AND	WRAPPING FUNCTIONS"""
#----------------------------------------------

# note:	you	must override both getWrap...()	functions or neither
def	getWrapX():
	"Can be	overridden to change whether the map wraps around in the X direction. Default is true"
	return True
def	getWrapY():
	"Can be	overridden to change whether the map wraps around in the Y direction. Default is false"
	return False

def	getTopLatitude():
	"Can be	overridden to change whether the the top latitude. Default is 90"
	initClimateSettings()
	cgc = CyGlobalContext()
	map = cgc.getMap()
	iH = map.getGridHeight()
	topLatitudeTemperature = getTileTemperature(0,iH)
	result = 30 - topLatitudeTemperature * 2
	return int(result)

def	getBottomLatitude():
	"Can be	overridden to change whether the the bottom	latitude. Default is -90"
	cgc = CyGlobalContext()
	map = cgc.getMap()
	iH = map.getGridHeight()
	topLatitudeTemperature = getTileTemperature(0,iH)
	result = - (30 - topLatitudeTemperature * 2)
	return int(result)

def	isBonusIgnoreLatitude():
	"Can be	overridden to determine	whether	bonus generation ignores latitude. Default is false"
	CyPythonMgr().allowDefaultImpl()